/* CSS main parser */
/* $Id: parser.c,v 1.92 2004/03/30 00:47:27 jonas Exp $ */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdlib.h>
#include <string.h>

#include "elinks.h"

#include "document/css/parser.h"
#include "document/css/property.h"
#include "document/css/scanner.h"
#include "document/css/stylesheet.h"
#include "document/css/value.h"
#include "document/html/parser.h"
#include "util/color.h"
#include "util/lists.h"
#include "util/error.h"
#include "util/memory.h"
#include "util/string.h"


void
css_parse_properties(struct list_head *props, struct scanner *scanner)
{
	assert(props && scanner);

	while (scanner_has_tokens(scanner)) {
		struct css_property_info *property_info = NULL;
		struct css_property *prop;
		struct scanner_token *token = get_scanner_token(scanner);
		int i;

		if (!token || token->type == '}') break;

		/* Extract property name. */

		if (token->type != CSS_TOKEN_IDENT
		    || !check_next_scanner_token(scanner, ':')) {
			/* Some use style="{ properties }" so we have to be
			 * check what to skip to. */
			if (token->type == '{') {
				skip_scanner_token(scanner);
			} else {
				skip_css_tokens(scanner, ';');
			}
			continue;
		}

		for (i = 0; css_property_info[i].name; i++) {
			struct css_property_info *info = &css_property_info[i];

			if (scanner_token_strlcasecmp(token, info->name, -1)) {
				property_info = info;
				break;
			}
		}

		/* Skip property name and separator and check for expression */
		if (!skip_css_tokens(scanner, ':')) {
			assert(!scanner_has_tokens(scanner));
			break;
		}

		if (!property_info) {
 			/* Unknown property, check the next one. */
 			goto ride_on;
 		}

		/* We might be on track of something, cook up the struct. */

		prop = mem_calloc(1, sizeof(struct css_property));
		if (!prop) {
			goto ride_on;
		}
		prop->type = property_info->type;
		prop->value_type = property_info->value_type;
		if (!css_parse_value(property_info, &prop->value, scanner)) {
			mem_free(prop);
			goto ride_on;
		}
		add_to_list(*props, prop);

		/* Maybe we have something else to go yet? */

ride_on:
		skip_css_tokens(scanner, ';');
	}
}


/* TODO: We should handle support for skipping blocks better like "{ { } }"
 * will be handled correctly. --jonas */
#define skip_css_block(scanner) \
	if (skip_css_tokens(scanner, '{')) skip_css_tokens(scanner, '}');


/* Atrules grammer:
 *
 * media_types:
 *	  <empty>
 *	| <ident>
 *	| media_types ',' <ident>
 *
 * atrule:
 * 	  '@charset' <string> ';'
 *	| '@import' <string> media_types ';'
 *	| '@import' <uri> media_types ';'
 *	| '@media' media_types '{' ruleset* '}'
 *	| '@page' <ident>? [':' <ident>]? '{' properties '}'
 *	| '@font-face' '{' properties '}'
 */
static void
css_parse_atrule(struct css_stylesheet *css, struct scanner *scanner)
{
	struct scanner_token *token = get_scanner_token(scanner);

	/* Skip skip skip that code */
	switch (token->type) {
		case CSS_TOKEN_AT_IMPORT:
			token = get_next_scanner_token(scanner);
			if (!token) break;

			if (token->type == CSS_TOKEN_STRING
			    || token->type == CSS_TOKEN_URL) {
				assert(css->import);
				css->import(css, token->string, token->length);
			}
			skip_css_tokens(scanner, ';');
			break;

		case CSS_TOKEN_AT_CHARSET:
			skip_css_tokens(scanner, ';');
			break;

		case CSS_TOKEN_AT_FONT_FACE:
		case CSS_TOKEN_AT_MEDIA:
		case CSS_TOKEN_AT_PAGE:
			skip_css_block(scanner);
			break;

		case CSS_TOKEN_AT_KEYWORD:
			/* TODO: Unkown @-rule so either skip til ';' or next block. */
			while (scanner_has_tokens(scanner)) {
				token = get_next_scanner_token(scanner);

				if (!token) break;

				if (token->type == ';') {
					skip_scanner_token(scanner);
					break;

				} else if (token->type == '{') {
					skip_css_block(scanner);
					break;
				}
			}
			break;
		default:
			INTERNAL("@-rule parser called without atrule.");
	}
}


struct selector_pkg {
	LIST_HEAD(struct selector_pkg);
	struct css_selector *selector;
};

/* Selector grammar:
 *
 * selector:
 *	  (element_name id? class? pseudo_class?)+
 *	  | '#' ('.' class)? (':' pseudo_class)?
 *	  | '.' class (':' pseudo_class)?
 *	  | ':' pseudo_class
 *
 * TODO: selector can currently only be simple element names, and element
 * chains are not supported yet.
 */
static void
css_parse_selector(struct css_stylesheet *css, struct scanner *scanner,
		   struct list_head *selectors)
{
	/* TODO: selector is (<element>)?([#:.]<ident>)?, not just <element>.
	 * And anyway we should have css_parse_selector(). --pasky */
	/* TODO: comma-separated list of simple selectors. */
	/* FIXME: element can be even '*' --pasky */

	while (scanner_has_tokens(scanner)) {
		struct scanner_token *token = get_scanner_token(scanner);
		struct scanner_token element;
		struct selector_pkg *pkg = NULL;
		struct css_selector *selector;

		assert(token);

		/* This is basicly check_css_precedence(token->type, ','))
		 * meaning check if the token precedence is less than or equal
		 * some ordinary token. But that would be too corny to write at
		 * this stage. */
		if (token->type == '{'
		    || token->type == '}'
		    || token->type == ';')
			break;

		/* Load up the selector */

		if (token->type != CSS_TOKEN_IDENT) {
			/* FIXME: Temporary fix for this weird CSS precedence
			 * thing. ')' has higher than ',' and it can cause
			 * problems when skipping here. The reason is for the
			 * function() parsing. Hmm... --jonas */
			if (!skip_css_tokens(scanner, ','))
				skip_scanner_token(scanner);
			continue;
		}

		/* Save it */
		memcpy(&element, token, sizeof(struct scanner_token));
		token = get_next_scanner_token(scanner);
		if (!token) return;

		/* Let's see if we will get anything else of this. */

		switch (token->type) {
			/* id */
			case CSS_TOKEN_HASH:
			case CSS_TOKEN_HEX_COLOR:
				skip_css_tokens(scanner, ',');
				continue;

			/* class */
			case '.':
			/* pseudo */
			case ':':
				skip_css_tokens(scanner, ',');
				continue;

			case ',':
			case '{':
				/* These are the only ones we are currently
				 * interested in. */
				break;

			default:
				skip_css_tokens(scanner, ',');
				continue;
		}

		/* Check if we have already encountered the selector */

		/* FIXME: This is totally broken because we have to do this _after_
		 * scanning for id/class/pseudo. --pasky */
		selector = get_css_selector(css, element.string, element.length);
		if (!selector) continue;

		pkg = mem_calloc(1, sizeof(struct selector_pkg));
		if (!pkg) continue;
		pkg->selector = selector;

		/* This is a temporary measure, so that overly specific selectors won't
		 * spoil our whole stylesheet until we support them, and we always
		 * favour the general selectors instead. Keep just the add_to_list()
		 * when we will start supporting the id/class/pseudo. */
		add_to_list(*selectors, pkg);

		/* Multiple elements hooked up to this ruleset. */
		if (token->type == ',') skip_scanner_token(scanner);
	}
}


/* Ruleset grammar:
 *
 * ruleset:
 *	  selector [ ',' selector ]* '{' properties '}'
 */
static void
css_parse_ruleset(struct css_stylesheet *css, struct scanner *scanner)
{
	INIT_LIST_HEAD(selectors);
	INIT_LIST_HEAD(properties);
	struct selector_pkg *pkg;

	css_parse_selector(css, scanner, &selectors);
	if (list_empty(selectors)
	    || !skip_css_tokens(scanner, '{')) {
		if (!list_empty(selectors)) free_list(selectors);
		skip_css_tokens(scanner, '}');
		return;
	}


	/* We don't handle the case where a property has already been added to
	 * a selector. That doesn't matter though, because the best one will be
	 * always the last one (FIXME: 'important!'), therefore the applier
	 * will take it last and it will have the "final" effect.
	 *
	 * So it's only a little waste and no real harm. The thing is, what do
	 * you do when you have 'background: #fff' and then 'background:
	 * x-repeat'? It would require yet another logic to handle merging of
	 * these etc and the induced overhead would in most cases mean more
	 * waste that having the property multiple times in a selector, I
	 * believe. --pasky */

	pkg = selectors.next;
	css_parse_properties(&properties, scanner);

	skip_css_tokens(scanner, '}');

	/* Mirror the properties to all the selectors. */
	foreach (pkg, selectors) {
		add_selector_properties(pkg->selector, &properties);
	}
	free_list(selectors);
	free_list(properties);
}


void
css_parse_stylesheet(struct css_stylesheet *css, unsigned char *string,
		     unsigned char *end)
{
	struct scanner scanner;

	init_scanner(&scanner, &css_scanner_info, string, end);

	while (scanner_has_tokens(&scanner)) {
		struct scanner_token *token = get_scanner_token(&scanner);

		assert(token);

		switch (token->type) {
		case CSS_TOKEN_AT_KEYWORD:
		case CSS_TOKEN_AT_CHARSET:
		case CSS_TOKEN_AT_FONT_FACE:
		case CSS_TOKEN_AT_IMPORT:
		case CSS_TOKEN_AT_MEDIA:
		case CSS_TOKEN_AT_PAGE:
			css_parse_atrule(css, &scanner);
			break;

		default:
			/* And WHAT ELSE could it be?! */
			css_parse_ruleset(css, &scanner);
		}
	}
}
